/* This file is part of guile-with-graphics, a graphic library for Guile

Copyright (C) 2008, 2009  Luca Saiu
Based on turtle, by Luca Saiu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include "sdl.h"
#include "debug.h"
#include <SDL.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdbool.h>
#include <math.h>
#include <assert.h>
#include "compiletimeparameters.h"
#include "color.h"
//#include "SDL/SDL_syswm.h"

/* The global, one and only window we use. Note how this indirection is quite
   inefficient and not needed; I just do that so that it's easier, in the future,
   to add support for more than one window at a time: */
game_window_t the_window = NULL;

/* Get the next event, if any, and return a pointer to it. A pointer to a static
   buffer is returned, which is overwritten by subsequent calls to game_get_event().
   The user must *not* free() the returned buffer. 
   NULL is returned when there are no pending events: */
SDL_Event* game_get_event(void){
  static SDL_Event event;
  
  /* Put pending events (if any) into the queue, so that they are visible by
     SDL_PollEvent(): */
  SDL_PumpEvents();
  
  /* Do we have anything in the event queue? If so copy the event into the
     static buffer, and return a pointer to it; otherwise return NULL: */
  if(SDL_PollEvent(&event) != 0)
    return &event;
  else
    return NULL;
}

/* Return the next event converted into an s-expression, or #f if no pending
   event exists: */
SCM game_get_event_guile(void){
  /* Create symbols, if we haven't already done that. scm_from_locale_symbol()
     is relatively slow, so we don't want to call it every time, but just at
     initialization, once per symbol.
     The following macros help to avoid duplicating some code. If you don't
     understand how it works but you're interested you can write me. */
#define GAME_MAKE_SYMBOL(C_NAME, SCHEME_NAME) \
  static SCM C_NAME ## _symbol; \
  if(! have_we_created_symbols) \
    C_NAME ## _symbol = scm_permanent_object(scm_from_locale_symbol( #SCHEME_NAME ));
#define GAME_MAKE_SYMBOL1(C_AND_SCHEME_NAME) \
  GAME_MAKE_SYMBOL(C_AND_SCHEME_NAME, C_AND_SCHEME_NAME)
  
  static bool have_we_created_symbols = false;
  GAME_MAKE_SYMBOL1(quit)
  GAME_MAKE_SYMBOL1(active)
  GAME_MAKE_SYMBOL1(expose)
  GAME_MAKE_SYMBOL1(resize)
  GAME_MAKE_SYMBOL(key_down, key-down)
  GAME_MAKE_SYMBOL(key_up, key-up)
  GAME_MAKE_SYMBOL1(left)
  GAME_MAKE_SYMBOL1(right)
  GAME_MAKE_SYMBOL1(up)
  GAME_MAKE_SYMBOL1(down)
  GAME_MAKE_SYMBOL1(backspace)
  GAME_MAKE_SYMBOL1(space)
  GAME_MAKE_SYMBOL1(enter)
  GAME_MAKE_SYMBOL1(escape)
  GAME_MAKE_SYMBOL1(tab)
  GAME_MAKE_SYMBOL1(unknown)
  GAME_MAKE_SYMBOL(left_control, left-control)
  GAME_MAKE_SYMBOL(left_alt, left-alt)
  GAME_MAKE_SYMBOL(left_shift, left-shift)
  GAME_MAKE_SYMBOL(right_control, right-control)
  GAME_MAKE_SYMBOL(right_alt, right-alt)
  GAME_MAKE_SYMBOL(right_shift, right-shift)
  GAME_MAKE_SYMBOL(left_meta, left-meta)
  GAME_MAKE_SYMBOL(left_super, left-super)
  GAME_MAKE_SYMBOL(right_meta, right-meta)
  GAME_MAKE_SYMBOL(right_super, right-super)
  GAME_MAKE_SYMBOL1(mode)
  GAME_MAKE_SYMBOL1(compose)
  GAME_MAKE_SYMBOL(caps_lock, caps-lock)
  GAME_MAKE_SYMBOL(num_lock, num-lock)
  GAME_MAKE_SYMBOL1(f1)
  GAME_MAKE_SYMBOL1(f2)
  GAME_MAKE_SYMBOL1(f3)
  GAME_MAKE_SYMBOL1(f4)
  GAME_MAKE_SYMBOL1(f5)
  GAME_MAKE_SYMBOL1(f6)
  GAME_MAKE_SYMBOL1(f7)
  GAME_MAKE_SYMBOL1(f8)
  GAME_MAKE_SYMBOL1(f9)
  GAME_MAKE_SYMBOL1(f10)
  GAME_MAKE_SYMBOL1(f11)  
  GAME_MAKE_SYMBOL1(f12)
  GAME_MAKE_SYMBOL1(home)
  GAME_MAKE_SYMBOL1(end)
  GAME_MAKE_SYMBOL1(delete)
  GAME_MAKE_SYMBOL1(insert)
  GAME_MAKE_SYMBOL(page_up, page-up)
  GAME_MAKE_SYMBOL(page_down, page-down)

  /* If we are here then the  initialization is complete: */
  have_we_created_symbols = true;
  
  /* Take the next SDL event, if any, and translate it: */
  SDL_Event *event = game_get_event();
  if(event == NULL) // no event: return #f
    return SCM_BOOL_F;
  else // convert the SDL event in the C struct into a Scheme s-expression:
    switch(event->type){
    case SDL_VIDEOEXPOSE:
      return scm_list_1(expose_symbol);
    case SDL_ACTIVEEVENT:
      return scm_list_2(active_symbol, event->active.gain ? SCM_BOOL_T : SCM_BOOL_F);
    case SDL_VIDEORESIZE:
      return scm_list_3(resize_symbol,
                        scm_from_int(event->resize.w),
                        scm_from_int(event->resize.h));
    case SDL_KEYDOWN: case SDL_KEYUP:{ // a key was pressed or de-pressed
      /* By default, let's turn the SDL key, which is encoded as an 8-bit unsigned
         integer in a way to be compatible with ASCII, into a Scheme character: */      
      SCM key = SCM_MAKE_CHAR(event->key.keysym.sym);
      /* There are some particular cases for which we don't have a suitable character
         or it's better to use a symbol with a clear name, so we use a Scheme symbol
         instead of a Scheme character: */
      switch(event->key.keysym.sym){
      case SDLK_UP: key = up_symbol; break;
      case SDLK_DOWN: key = down_symbol; break;
      case SDLK_RIGHT: key = right_symbol; break;
      case SDLK_LEFT: key = left_symbol; break;
      case SDLK_BACKSPACE: key = backspace_symbol; break; 
      case SDLK_SPACE: key = space_symbol; break;
      case SDLK_RETURN: key = enter_symbol; break;
      case SDLK_TAB: key = tab_symbol; break;
      case SDLK_ESCAPE: key = escape_symbol; break;
      case SDLK_LCTRL: key = left_control_symbol; break;
      case SDLK_LALT: key = left_alt_symbol; break;
      case SDLK_LSHIFT: key = left_shift_symbol; break;
      case SDLK_LMETA: key = left_meta_symbol; break;
      case SDLK_LSUPER: key = left_super_symbol; break;
      case SDLK_RCTRL: key = right_control_symbol; break;
      case SDLK_RALT: key = right_alt_symbol; break;
      case SDLK_RSHIFT: key = right_shift_symbol; break;
      case SDLK_RMETA: key = right_meta_symbol; break;
      case SDLK_RSUPER: key = right_super_symbol; break;
      case SDLK_MODE: key = mode_symbol; break;
      case SDLK_COMPOSE: key = compose_symbol; break;
      case SDLK_CAPSLOCK: key = caps_lock_symbol; break;
      case SDLK_NUMLOCK: key = num_lock_symbol; break;
      case SDLK_F1: key = f1_symbol; break;
      case SDLK_F2: key = f2_symbol; break;
      case SDLK_F3: key = f3_symbol; break;
      case SDLK_F4: key = f4_symbol; break;
      case SDLK_F5: key = f5_symbol; break;
      case SDLK_F6: key = f6_symbol; break;
      case SDLK_F7: key = f7_symbol; break;
      case SDLK_F8: key = f8_symbol; break;
      case SDLK_F9: key = f9_symbol; break;
      case SDLK_F10: key = f10_symbol; break;
      case SDLK_F11: key = f11_symbol; break;
      case SDLK_F12: key = f12_symbol; break;
      case SDLK_HOME: key = home_symbol; break;
      case SDLK_END: key = end_symbol; break;
      case SDLK_INSERT: key = insert_symbol; break;
      case SDLK_DELETE: key = delete_symbol; break;
      case SDLK_PAGEUP: key = page_up_symbol; break;
      case SDLK_PAGEDOWN: key = page_down_symbol; break;
      } // inner switch
      return scm_list_2((event->type == SDL_KEYDOWN) ? key_down_symbol : key_up_symbol, key);
    }
    case SDL_QUIT:
      return scm_list_1(quit_symbol);
    default:
      //return scm_list_1(unknown_symbol);
      return SCM_BOOL_F;
    } // switch
}

/* Initialize SDL support, but not the graphic subsystem (yet): */
static void game_initialize_SDL(void){
  /* Initialize SDL, or die: */
  if(SDL_Init(0/* SDL_INIT_EVENTTHREAD */) < 0)
    game_fatal("Couldn't initialize SDL video support: %s\n", SDL_GetError());
  
  /* Execute SDL_Quit() when we exit (*not* now): */
  atexit(SDL_Quit);
}

/* Make a new window of the given size and return it, or fail with a 
   fatal error: */
static game_window_t game_make_window(const char *window_title, const int width, const int height){
  /* Initialize the SDL video subsystem, thus making a window; notice that we
     can only make one window at the time in this way: */
  SDL_InitSubSystem(SDL_INIT_VIDEO);
  
  /* Set the window title: */
  SDL_WM_SetCaption(window_title, NULL /* no icon */);
  
  /* Make the window "always on top": */
  // To do: I'd like to do this, but it's not supported by SDL. I could do it
  // via X and the syswm interface, but it's ugly. Any ideas?
  
  /* Allocate space for the result: */
  game_window_t result = (game_window_t)
    game_malloc(sizeof(struct game_window));
  
  /* Make the window, and fill the structure: */
  const int preferred_depth = 0; // any depth will do
  result->default_surface =
    SDL_SetVideoMode(width, height, preferred_depth, SDL_SWSURFACE);
  if(result->default_surface == NULL)
    game_fatal("could not make a %ix%i window (%i bpp): %s\n",
                 width, height, preferred_depth,
                 SDL_GetError());
  
  /* Make another surface, which may be needed for some applications: */
  result->scratch_surface =
    SDL_CreateRGBSurface(SDL_SWSURFACE, width, height,
                         result->default_surface->format->BitsPerPixel,
                         0, 0, 0, 0);
  if(result->scratch_surface == NULL)
    game_fatal("could not make the scratch surface");
  
  /* We're done, return our flashy new window: */
  return result;
}

int game_get_window_width(void){
  return the_window->default_surface->w;
}

int game_get_window_height(void){
  return the_window->default_surface->h;
}

void game_initialize_graphics(const char *window_title, const int width, const int height){
  /* Fail if we have already initialized: */
  if(the_window != NULL)
    game_fatal("a window already exists");
  
  /* Initialize SDL and make the window: */
  game_initialize_SDL();
  the_window = game_make_window(window_title, width, height);
  
  /* Also clear the window, sometimes it's required (depending on the
     video mode and/or the graphic driver and/or the X server, I think): */
  game_clear_window(game_color_black);
}

/* Destroy the given window, and deallocate the structure: */
static void game_destroy_window(const game_window_t window){
  /* Destroy all surfaces: */
  SDL_FreeSurface(the_window->scratch_surface);
  
  /* Finalize the SDL video subsystem, thus destroying the window: */
  SDL_QuitSubSystem(SDL_INIT_VIDEO);
  
  /* Free the buffer: */
  game_free(window);
}

void game_finalize_graphics(void){
  /* Fail if we have not finalized: */
  if(the_window == NULL)
    game_fatal("finalizing twice?");
  
  /* Destroy the window and finalize all SDL support: */
  game_destroy_window(the_window);
  the_window = NULL;
  SDL_Quit();
}

/* This is the same as lock_surface(), but can work on any surface, not
   only the main one: */
static void lock_the_given_surface(SDL_Surface *surface){
  /* If the surface requires locking, and it's now locked, wait until
     it's freed: */
  if(SDL_MUSTLOCK(surface))
    while(SDL_LockSurface(surface) < 0)
      /* Do nothing */;
}

/* This is the same as unlock_surface(), but can work on any surface, not
   only the main one: */
static void unlock_the_given_surface(SDL_Surface *surface){
  /* If the surface requires locking (assuming that *we* are holding its
     lock) then unlock it: */
  if(SDL_MUSTLOCK(surface))
    SDL_UnlockSurface(surface);
}

void game_lock_surface(void){
  lock_the_given_surface(the_window->default_surface);
}

void game_unlock_surface(void){
  unlock_the_given_surface(the_window->default_surface);
}

/* Draw a pixel in the given depth on the given surface, without locking: */
static void game_draw_pixel_8bpp(SDL_Surface *surface,
                                 int x, int y, const Uint32 color){
  Uint8 *bufp = (Uint8 *)surface->pixels + y*surface->pitch + x;
  *bufp = color;
}
static void game_draw_pixel_15or16bpp(SDL_Surface *surface,
                                      int x, int y, const Uint32 color){
  Uint16 *bufp = (Uint16 *)surface->pixels + y*surface->pitch/2 + x;
  *bufp = color;
}
static void game_draw_pixel_24bpp(SDL_Surface *surface,
                                  int x, int y,
                                  Uint8 r, Uint8 g, Uint8 b){
  Uint8 *bufp =(Uint8 *)surface->pixels + y*surface->pitch + x;
  *(bufp+surface->format->Rshift/8) = r;
  *(bufp+surface->format->Gshift/8) = g;
  *(bufp+surface->format->Bshift/8) = b;
}
static void game_draw_pixel_32bpp(SDL_Surface *surface,
                                  int x, int y, const Uint32 color){
  Uint32 *bufp = (Uint32 *)surface->pixels + y*surface->pitch/4 + x;
  *bufp = color;
}

void game_refresh_window(void){
  SDL_UpdateRect(the_window->default_surface, 0, 0, 0, 0);
}

void game_refresh_rectangle(const int x0_, const int y0_,
                            const int width_, const int height_){
  /* If either width or height are negative then retry with the *same*
     rectangle, but described so that width and height are non-negative: */
  if(width_ < 0)
    return game_refresh_rectangle(x0_ + width_, y0_, - width_, height_);
  if(height_ < 0)
    return game_refresh_rectangle(x0_, y0_ + height_, width_, - height_);
  
  /* Clip coordinates: */
  int x0 = x0_, y0 = y0_, width = width_, height = height_;
  if(x0_ < 0){
    x0 = 0; width += x0_;
  }
  if(y0_ < 0){
    y0 = 0; height += y0_;
  }
  const int x2 = x0 + width;
  const int y2 = y0 + height;
  if(x2 >= game_get_window_width())
    width = game_get_window_width() - x0;
  if(y2 >= game_get_window_height())
    height = game_get_window_height() - y0;
  
  /* After this clipping, if either the width or height is negative then
     the rectangle is out of the window; and if either is zero then the
     rectangle surface is zero: */
  if((width <= 0) || (height <= 0))
    return;
  
  /* Do nothing if the rectangle to refresh is out of the window (now we can
     be sure than width and height are positive): */
  if((x0 >= game_get_window_width()) || (y0 >= game_get_window_height()))
    return;

  //printf("[Refreshing the rectangle %i,%i+%i+%i\n", x0, y0, width, height);
  SDL_UpdateRect(the_window->default_surface, x0, y0, width, height);
}

/* Return true iff the pixel with the given coordinates in in the window: */
static bool is_within_bounds(const game_window_t window, const int x, const int y){
  return ((x >= 0) &&
          (y >= 0) &&
          (x < window->default_surface->w) &&
          (y < window->default_surface->h));
}

/* Note that no bound checks are performed: the coordinates are assumed to
   be within the given surface: */
static void game_draw_pixel_on_surface(SDL_Surface *surface,
                                       const int x, const int y,
                                       const game_color_t color){
  /* Do nothing if the pixel is out of the window: */
  if(! is_within_bounds(the_window, x, y))
    return;
  
  /* Actually draw the pixel; the exact way of doing it depends on the
     color depth: */
  switch(surface->format->BytesPerPixel){ // *bytes*, not bits, per pixel
  case 1:
    game_draw_pixel_8bpp(surface, x, y, color->color_as_uint32);
    break;
  case 2:
    game_draw_pixel_15or16bpp(surface, x, y, color->color_as_uint32);
    break;
  case 3:
    game_draw_pixel_24bpp(surface, x, y,
                          color->red_as_uint8, color->green_as_uint8, color->blue_as_uint8);
    break;    
  case 4:
    game_draw_pixel_32bpp(surface, x, y, color->color_as_uint32);
    break;
  default:
    game_fatal("I have no idea about how to draw at %i bytes per pixel",
               surface->format->BytesPerPixel);
  } // switch
}

/* Destructively swap two variables. Note that A and B are supposed to
   be variables, not other kinds of expressions. Explicit note for some
   students who may not be expert with C: this is a *macro*, not a
   function! You *can't* do this with a function, unless you use pointers.
   C is a call-by-value language, like Scheme.
   If you don't understand this explaination and you're interested of
   course you can ask me. */
#define SWAP(A, B) \
  { const typeof(A) temporary_ = A; \
    A = B; \
    B = temporary_; }

void game_draw_pixel(const int x, const int y, const game_color_t color){
  game_draw_pixel_on_surface(the_window->default_surface, x, y, color);
}

void game_draw_line(const int x1_, const int y1_,
                    const int x2_, const int y2_,
                    const game_color_t color){
  /* In the following I assume the width to be positive; if it's
     not, then I *make* it positive by exchanging the first point with
     the last point. Instead y1 and y2 may be in any order, but the
     height is always >= 0. I didn't find a cleaner solution which
     was simple enough. */
  int x1 = x1_, y1 = y1_, x2 = x2_, y2 = y2_;
  if(x1 > x2){
    SWAP(x1, x2);
    SWAP(y1, y2);
  } // if
  const int width = x2 - x1;
  const int min_y = (y1 < y2) ? y1 : y2;
  const int max_y = (y1 > y2) ? y1 : y2;
  const int height = max_y - min_y;
  const int y_step = (y1 < y2) ? 1 : -1;
  /* Ok, now we're ready to do the actual drawing. */
  
  /* Draw the line, pixel by pixel: */
  if(width != 0){ // is the line vertical?
    /* No, it's not vertical, so we can make this division: */
    const float inclination = ((float)(y2_ - y1_)) / ((float)(x2_ - x1_));
    
    /* I draw the line in two passes, so that also very inclined lines
       are shown correctly without "holes": in the first pass I compute
       the y coordinate given the x, and in the second pass I do the
       opposite: */
    
    /* For each x compute the corresponding y and draw the pixel: */
    float delta_x;
    for(delta_x = 0; delta_x <= width; delta_x ++){
      const int delta_y = (int)(delta_x * inclination);
      game_draw_pixel(x1 + delta_x, y1 + delta_y, color);
    } // for

    /* For each y compute the corresponding x and draw the pixel: */
    const float inverse_inclination = 1.0 / inclination;
    float delta_y;
    for(delta_y = 0; delta_y <= height; delta_y ++){
      const int delta_x = (int)(delta_y * inverse_inclination);
      game_draw_pixel(x1_ + delta_x, y1_ + delta_y, color);
    } // for
  }
  else{
    /* Yes, the line is vertical: just draw pixel per value of y. Only
       one pass is needed in this case: */
    int y;
    for(y = min_y; y <= max_y; y ++)
      game_draw_pixel(x1, y, color);
  } // outer else
}

void game_draw_rectangle(const int x0, const int y0,
                         const int width, const int height,
                         const game_color_t color){
  /* If either width or height are negative then draw the *same* rectangle,
     but described so that width and height are non-negative: */
  if(width < 0)
    return game_draw_rectangle(x0 + width, y0, - width, height, color);
  if(height < 0)
    return game_draw_rectangle(x0, y0 + height, width, - height, color);
  
  /* If we have arrived here then we're sure than width and height are
     non-negative. */
  
  /* Draw every pixel: */
  int delta_x, delta_y;
  for(delta_x = 0; delta_x < width; delta_x ++)
    for(delta_y = 0; delta_y < height; delta_y ++)
      game_draw_pixel(x0 + delta_x, y0 + delta_y, color);
}

void game_clear_window(const game_color_t color){
  /* Fill a rectangle; by the definition of this function when the
     rect is NULL the whole window is filled: */
  SDL_FillRect(the_window->default_surface, NULL, color->color_as_uint32);
}

Uint32 game_get_milliseconds_since_startup(void){
  return SDL_GetTicks();
}
