/* This file is part of guile-with-graphics, a graphic library for Guile

Copyright (C) 2008, 2009  Luca Saiu
Based on turtle, by Luca Saiu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#ifndef SDL_H_
#define SDL_H_

#include <SDL.h>
#include <libguile.h>
#include "color.h"

/* The data structure internally used for representing the state of the window: */
struct game_window{
  SDL_Surface *default_surface; // this is the surface mapped to the window
  SDL_Surface *scratch_surface; // this is another surface, of the same size
                                // of the main window
}; // struct
/* We always refer a window via a pointer to a malloc()ed buffer. Notice that
   we currently support only *one* window at a time: */
typedef struct game_window* game_window_t;

/* A pointer to the one and only window, initially NULL: */
extern game_window_t the_window;

/* Get the current window size: */
int game_get_window_width(void) __attribute__(( pure ));
int game_get_window_height(void) __attribute__(( pure ));

/* Initialize the graphics subsystem, making a new window of the given size,
   or fail with a fatal error: */
void game_initialize_graphics(const char *window_title, const int width, const int height);

/* Finalize the graphics subsystem: */
void game_finalize_graphics(void);

/* Lock the SDL surface. This must be explicitly called once, before starting
   to draw: */
void game_lock_surface(void);

/* Lock the SDL surface. This must be explicitly called once, after having
   finished to draw, before refreshing: */
void game_unlock_surface(void);

/* Draw a pixel on the window, at the given coordinates, with the given color.
   The window is *not* refreshed by this: the user must explicitly refresh a
   rectangle, or the whole window, when needed. */
void game_draw_pixel(const int x, const int y,
                     const game_color_t color);

/* Draw a line on the window, at the given position, with the given color,
   and show it immediately. Both extreme points are included. Again, no
   implicit refresh. */
void game_draw_line(const int x1, const int y1, // position of the first point
                    const int x2, const int y2, // position of the last point
                    const game_color_t color);

/* Draw a rectangle on the window, at the given position, with the given size,
   with the given color, Again, no implicit refresh. */
void game_draw_rectangle(const int x0, const int y0, // position of the first point
                         const int width, const int height,
                         const game_color_t color);

/* Clear the whole window by overdrawing with the given color. No refresh. */
void game_clear_window(const game_color_t color);

/* Refresh a rectangle of screen, so that any change within the rectangle
   becomes visible. This is expensive, so it's best to refresh only the smallest
   possible rectangle, and only when needed: */
void game_refresh_rectangle(const int x0, const int y0,
                            const int width, const int height);

/* Refresh the whole window, so that any change becomes visible: */
void game_refresh_window(void);

/* Get the next event, if any, and return a pointer to it. A pointer to a static
   buffer is returned, which is overwritten by subsequent calls to game_get_event().
   The user must *not* free() the returned buffer. 
   NULL is returned when there are no pending events: */
SDL_Event* game_get_event(void);

/* Return the next event converted into an s-expression, or #f if no pending
   event exists: */
SCM game_get_event_guile(void);

Uint32 game_get_milliseconds_since_startup(void);

#endif // #ifndef SDL_H_
